(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * Bank.thy --- an BirthdayBook inspired, elementary
 *              example for sequence testing based on a 
 *              deterministic specification.
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2009-15 Univ. Paris-Sud, France
 *               2009-15 Achim D. Brucker, Germany
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

theory  MyKeOS
imports 
   (* Testing *) 
  (*  "../../../src/codegen_fsharp/Code_Integer_Fsharp" *)
   "../../../src/Testing"
  "../../../src/codegen_gdb/Code_gdb_script"
  Interleaving
    
begin

declare [[testgen_profiling]]

text{* This is a fun-operating system --- closely following the Bank example ---
intended to explain the principles of symbolic execution used in our PikeOS study.

Moreover, in this scenario, we assume that the system under test is
deterministic.
 *}

text{* The state of a thread (belonging to a task, \ie{} a Unix/PosiX like ``process'' 
just modeled by a map from task-id/thread-id information to the number of a resource
(a communication channel descriptor, for example) that was allocated to a thread. *}
type_synonym task_id = int

type_synonym thread_id = int

type_synonym thread_local_var_tab = "(task_id \<times> thread_id) \<rightharpoonup> int"


subsubsection{* Operation definitions *}

text {* A standard, JML or OCL or VCC like interface specification
might look like:

\begin{verbatim}

Init:  forall (c,no) : dom(data_base::thread_local_var_tab). data_base(c,no)>=0

op alloc (c : task_id, no : thread_id, amount:nat) : unit
pre  (c,no) : dom(data_base)
post data_base'=data_base[(c,no) := data_base(c,no) + amount]

op release(c : task_id, no : thread_id, amount:nat) : unit
pre  (c,no) : dom(data_base) and data_base(c,no) >= amount
post data_base'=data_base[(c,no) := data_base(c,no) - amount]

op status (c : task_id, no : thread_id) : int
pre  (c,no) : dom(data_base)
post data_base'=data_base and result = data_base(c,no)


\end{verbatim}

*}



text{* Interface normalization turns this interface into the following input type: *}

datatype in_c = alloc  task_id thread_id nat
              | release task_id thread_id nat
              | status  task_id thread_id
              
typ "MyKeOS.in_c"

datatype out_c = alloc_ok | release_ok | status_ok nat 

fun    precond :: "thread_local_var_tab \<Rightarrow> in_c \<Rightarrow> bool"
where "precond \<sigma> (alloc c no m) = ((c,no) \<in> dom \<sigma>)"
    | "precond \<sigma> (status c no) = ((c,no) \<in> dom \<sigma>)"
    | "precond \<sigma> (release c no m) = ((c,no) \<in> dom \<sigma> \<and> (int m) \<le> the(\<sigma>(c,no)))"

fun    postcond :: "in_c \<Rightarrow> thread_local_var_tab \<Rightarrow> (out_c \<times> thread_local_var_tab) set"
where "postcond (alloc c no m) \<sigma> =
         { (n,\<sigma>'). (n = alloc_ok \<and> \<sigma>'=\<sigma>((c,no)\<mapsto> the(\<sigma>(c,no)) + int m))}"
    | "postcond (status c no) \<sigma> =
         { (n,\<sigma>'). (\<sigma>=\<sigma>' \<and> (\<exists> x. status_ok x = n \<and> x = nat(the(\<sigma>(c,no)))))}"
    | "postcond (release c no m) \<sigma> =
         { (n,\<sigma>'). (n = release_ok \<and> \<sigma>'=\<sigma>((c,no)\<mapsto> the(\<sigma>(c,no)) - int m))}"




subsubsection{* Constructing an Abstract Program *}

text{* Using the Operators \verb+impl+ and \verb+strong_impl+, we can 
synthesize an abstract program right away from the specification, i.e. the pair 
of pre- and  postcondition defined above. Since this program is even 
deterministic, we derive a set of symbolic execution rules used in the test case 
generation process which will produce symbolic results against which the PUT can 
be compared in the test driver. *}
(*
definition implementable :: "['\<sigma> \<Rightarrow> '\<iota> \<Rightarrow> bool,'\<iota> \<Rightarrow> ('o,'\<sigma>)MON_SB] \<Rightarrow> bool"
where     "implementable pre post =(\<forall> \<sigma> \<iota>. pre \<sigma> \<iota> \<longrightarrow>(\<exists> out \<sigma>'. (out,\<sigma>') \<in> post \<iota> \<sigma> ))"
*)

lemma precond_postcond_implementable: 
     "implementable precond postcond"
apply(auto simp: implementable_def)
apply(case_tac "\<iota>", simp_all)
done

text{* Based on this machinery, it is now possible to construct the system model as the 
       canonical completion of the (functional) specification consisting of pre-  and 
       post-conditions*}

definition "SYS = (strong_impl precond postcond)"

lemma SYS_is_strong_impl : "is_strong_impl precond postcond SYS"
by(simp add: SYS_def is_strong_impl)

thm SYS_is_strong_impl[simplified is_strong_impl_def,THEN spec,of "(alloc c no m)", simplified, standard]
(*automatic, but logically not stropng enough *)

subsubsection{* Proving Symbolic Execution Rules for the  Abstractly Program *}
text{* The following lemmas reveal that this "constructed" program is actually
(due to determinism of the spec): *}

lemma Eps_split_eq' : "(SOME (x', y'). x'= x  \<and> y'= y) = (SOME (x', y'). x = x' \<and> y = y')"
by(rule arg_cong[of _ _ "Eps"], auto)

interpretation alloc : efsm_det 
                      "precond" "postcond" "SYS" "(alloc c no m)" "\<lambda>_. alloc_ok" 
                      "\<lambda> \<sigma>. \<sigma>((c, no) \<mapsto>  (the(\<sigma>(c, no)) + int m))" "\<lambda> \<sigma>. ((c, no) \<in> dom \<sigma>)"
               by unfold_locales (auto simp: SYS_def Eps_split_eq')

interpretation release : efsm_det 
                        "precond" "postcond" "SYS" "(release c no m)" "\<lambda>_. release_ok" 
                         "\<lambda> \<sigma>. \<sigma>((c, no) \<mapsto>  (the(\<sigma>(c, no))-int m))" 
                         "\<lambda> \<sigma>.((c, no)\<in>dom \<sigma>) \<and> (int m)\<le>the(\<sigma>(c,no))"
               by unfold_locales (auto simp: SYS_def Eps_split_eq')

interpretation status : efsm_det 
                       "precond" "postcond" "SYS" "(status c no)" 
                       "\<lambda>\<sigma>. (status_ok (nat(the(\<sigma>(c, no)))))" 
                       "\<lambda> \<sigma>. \<sigma>" "\<lambda> \<sigma>. ((c, no) \<in> dom \<sigma>)"
               by unfold_locales (auto simp: SYS_def Eps_split_eq')




subsection{* Setup *}

text{* Now we close the theory of symbolic execution by \emph{exluding} elementary rewrite
steps on @{term mbind}, \ie{} the rules @{thm mbind.simps(1)} @{thm mbind.simps(2)} *}

declare mbind.simps(1) [simp del]
        mbind.simps(2) [simp del]

text{* Here comes an interesting detail revealing the power of the
approach: The generated sequences still respect the preconditions
imposed by the specification  - in this case, where we are talking about
a task_id for which a defined account exists and for which we will never
produce traces in which we release more money than available on it. *}

subsubsection{* Restricting the Test-Space by Test Purposes *}
text{* We introduce a constraint on the input sequence, in order to limit the test-space
       a little and eliminate logically possible, but irrelevant test-sequences for a specific
       test-purpose. In this case, we narrow down on test-sequences concerning a specific
       task_id @{term c} with a specific bank-account number @{term no}.
       
       We make the (in this case implicit, but as constraint explicitly stated) test hypothesis,
       that the @{term SUT} is correct if it behaves correct for a single task_id.
       This boils down to the assumption that they are implemented as atomic transactions
       and interleaved processing does not interfere with a single thread. *}

fun test_purpose :: "[(task_id \<times> thread_id) list, in_c list] \<Rightarrow> bool"
where
  "test_purpose ((c,no)#R) [status c' no'] = ((c=c' \<and> no=no') \<or> test_purpose R [status c' no'])" 
| "test_purpose ((c,no)#R) ((alloc c' no' m)#S) = ((c=c' \<and> no=no' \<and> test_purpose ((c,no)#R) S)
                                                   \<or> test_purpose R ((alloc c' no' m)#S))"
| "test_purpose ((c,no)#R) ((release c' no' m)#S) = ((c=c' \<and> no=no' \<and> test_purpose ((c,no)#R) S)
                                                   \<or> test_purpose R ((release c' no' m)#S))"
| "test_purpose _ _ = False"

(* Definition buggy --- the R list must be separated into two parts, but one con not throw 
   away elements) *)

lemma [simp] : "test_purpose [] a = False" by simp
lemma [simp] : "test_purpose r [] = False" by simp

lemma [simp] : "test_purpose ((c,no)#R) [a] = ((a = status c no) \<or> test_purpose R [a])"
proof (induct "R")
  case Nil show ?case by(cases a, auto)
next
  case (Cons a' R') then show ?case 
       apply(cases a, simp_all)
       apply(cases "a'", simp)
       apply(cases "a'", simp)
       apply(cases "a'", simp)
       apply(rename_tac "int1" "int2" "aa" "b")
       apply(case_tac "c = int1 \<and> no = int2", auto)
       done
qed

lemma [simp] : "R\<noteq>[] \<Longrightarrow> test_purpose [(c,no),(c',no')] (a#R) = 
                          (((\<exists>m. a = (alloc c no m)) \<or> 
                            (\<exists>m. a = (release c no m)) \<or> 
                            (\<exists>m. a = (alloc c' no' m)) \<or> 
                            (\<exists>m. a = (release c' no' m))) 
                            \<and> test_purpose [(c,no),(c',no')] R) "
apply(simp add: List.neq_Nil_conv, elim exE,simp)
apply(cases a, auto) 
apply(case_tac y, auto) 
apply(case_tac ys, auto) 
apply(case_tac y, auto) 
apply(case_tac ys, auto) 

sorry


subsubsection{* The TestGen Setup*}
text{* The default configuration of \verb$gen_test_cases$ does \emph{not} descend into 
       sub-type expressions of type constructors (since this is not always desirable, the choice 
       for the default had been for "non-descent"). This case is relevant here since 
       @{typ "MyKeOS.in_c list"} has just this structure but we need ways to explore the input sequence
       type further. Thus, we need configure, for all test cases, and derivation descendants of
       the relusting clauses during splitting, again splitting for all parameters of 
       input type @{typ "MyKeOS.in_c"}: *}

set_pre_safe_tac{*
  (fn ctxt => TestGen.ALLCASES(
                  TestGen.CLOSURE (
                      TestGen.case_tac_typ ctxt ["MyKeOS.in_c"]))) 
*}

subsubsection{* The Scenario*}
 
text{* We construct test-sequences for a concrete task_id (implicitely assuming 
that interleaving actions with other task_ids will not influence the system behaviour.
In order to prevent \testgen to perform case-splits over names --- i.e. list of 
characters --- we define it as constant.
*}
(* fixing the init-space a little *)
definition tid\<^sub>0 :: "task_id" where "tid\<^sub>0 = 0"
definition tid\<^sub>1 :: "task_id" where "tid\<^sub>1 = 1"

definition thid\<^sub>0 :: "thread_id" where "thid\<^sub>0 = 4"
definition thid\<^sub>1 :: "task_id" where "thid\<^sub>1 = 6"
(* PUT must be a constant for coding conventions ... *)

consts PUT :: "(in_c \<Rightarrow>(out_c, '\<sigma>)MON\<^sub>S\<^sub>E)"

declare[[goals_limit = 500]]
(* declare[[testgen_trace]] *)

subsubsection{* Making my own test-data generation --- temporarily *}
lemma HH : "(A \<and> (A \<longrightarrow> B)) = (A \<and>  B)" by auto

lemma P1 : "PO (((c\<^sub>0, 1) \<in> dom (empty((c\<^sub>0, 1):=Some 10)) \<longrightarrow>
         [status_ok (nat (the ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1))))] = [status_ok (nat (the ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1))))] \<and>
         (empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1) = Some 10) \<and>
        ((c\<^sub>0, 1) \<notin> dom (empty((c\<^sub>0, 1):=Some 10)) \<longrightarrow> [] = [status_ok (nat (the ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1))))] \<and> (empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1) = Some 10))"
by(auto simp: TestGen.PO_def)


lemma P1' :  "PO (((c\<^sub>0, 1) \<in> dom (empty((c\<^sub>0, 1):=Some 10)) \<longrightarrow>
         (empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1) = Some 10 \<and>
         [status_ok (nat (the ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1))))] = [status_ok (nat (the ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1))))]) \<and>
        ((c\<^sub>0, 1) \<notin> dom (empty((c\<^sub>0, 1):=Some 10)) \<longrightarrow>
         (empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1) = Some 10 \<and> [] = [status_ok (nat (the ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1))))]))"
by(auto simp: TestGen.PO_def)

lemma P1'' : "PO ((empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1) = Some 10 \<and> (empty((c\<^sub>0, 1):=Some 10)) (c\<^sub>0, 1) = Some 10)"
by(auto simp: TestGen.PO_def)


lemma P2 : "PO (((1::int) = 0 \<longrightarrow>
          ((1::int) = 0 \<longrightarrow>
           ((c\<^sub>0, 0) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [alloc_ok, alloc_ok,
             status_ok
              (nat (the ([(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0)) + int (Suc(Suc(Suc(Suc(0))))) + int(Suc(Suc 0))))] =
            [alloc_ok, alloc_ok, status_ok 12] \<and>
            [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5) \<and>
           ((c\<^sub>0, 0) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5)) \<and>
          ((c\<^sub>0, 1::int) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
           ((c\<^sub>0, 0) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [alloc_ok, alloc_ok,
             status_ok
              (nat (the ([(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0)) + int (Suc(Suc(Suc(Suc(0))))) + int(Suc(Suc 0))))] =
            [alloc_ok, alloc_ok, status_ok 12] \<and>
            [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5) \<and>
           ((c\<^sub>0, 0) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5)) \<and>
          ((1::int) \<noteq> 0 \<and> (c\<^sub>0, 1::int) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
           ((c\<^sub>0, 0) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [alloc_ok] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5) \<and>
           ((c\<^sub>0, 0) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5))) \<and>
         ((1::int) \<noteq> 0 \<longrightarrow>
          ((1::int) = 0 \<longrightarrow>
           ((c\<^sub>0, 0) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [alloc_ok, alloc_ok,
             status_ok (nat (the ([(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 1::int)) + int(Suc(Suc 0))))] =
            [alloc_ok, alloc_ok, status_ok 12] \<and>
            [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5) \<and>
           ((c\<^sub>0, 0) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5)) \<and>
          ((c\<^sub>0, 1::int) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
           ((c\<^sub>0, 0) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [alloc_ok, alloc_ok,
             status_ok (nat (the ([(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 1::int)) + int(Suc(Suc 0))))] =
            [alloc_ok, alloc_ok, status_ok 12] \<and>
            [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5) \<and>
           ((c\<^sub>0, 0) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5)) \<and>
          ((1::int) \<noteq> 0 \<and> (c\<^sub>0, 1::int) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
           ((c\<^sub>0, 0) \<in> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [alloc_ok] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5) \<and>
           ((c\<^sub>0, 0) \<notin> dom [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] \<longrightarrow>
            [] = [alloc_ok, alloc_ok, status_ok 12] \<and> [(c\<^sub>0, 1::int) \<mapsto> 10, (c\<^sub>0, 0) \<mapsto> 5] (c\<^sub>0, 0) = Some 5))))"
by(simp add: PO_def)

lemma P2'' : "PO (([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1) = Some 10 \<and>
              [alloc_ok, alloc_ok, status_ok (nat (10 + int 4 + int 7))] = [alloc_ok, alloc_ok, status_ok (nat (10 + int 4 + int 7))] \<and>
              ([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1) = Some 10)"
by(auto simp: TestGen.PO_def)


lemma P3 : "PO (((c\<^sub>0,1) \<in> dom([(c\<^sub>0, 1) \<mapsto> 10]) \<and> int 5 \<le> 10 \<longrightarrow>
         [release_ok, status_ok (nat (the (([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1)) - int 5))] =
         [release_ok, status_ok (nat (the (([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1)) - int 5))] \<and>
         ([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1) = Some 10) \<and>
        (((c\<^sub>0, 1) \<in> dom ([(c\<^sub>0, 1) \<mapsto> 10]) \<longrightarrow> \<not> int 5 \<le> 10) \<longrightarrow>
         [] = [release_ok, status_ok (nat (the (([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1)) - int 5))] \<and> ([(c\<^sub>0, 1) \<mapsto> 10]) (c\<^sub>0, 1) = Some 10))"
by(auto simp: TestGen.PO_def)


subsubsection{* Small, rewriting based Scenarios including standard code-generation *}

text{* Exists in two formats : General Fail-Safe Tests (which allows for scenarios 
with normal \emph{and} exceptional behaviour; and Fail-Stop Tests, which generates Tests 
only for normal behaviour and correspond to inclusion test refinement. *}

text{* In the following, we discuss a test-scenario with error-abort semantics; \ie{} in each
test-case, a sequence may be chosen (by the test data selection) where the task_id has  several
accounts. . . *}
test_spec test_status:
assumes account_def   : "(c\<^sub>0,no) \<in> dom \<sigma>\<^sub>0" 
and     store_finite  : "\<sigma>\<^sub>0 = map_of T" 
and     test_purpose  : "test_purpose [(c\<^sub>0,no),(c\<^sub>0,no')] S"
and     sym_exec_spec :
       " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>a\<^sub>v\<^sub>e S SYS; return (s = x))"
shows  " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>a\<^sub>v\<^sub>e S PUT; return (s = x))"

txt{* Prelude: Massage of the test-theorem --- representing the assumptions of the test explicitely
      in HOL and blocking @{term x} from beeing case-splitted (which complicates the process).*}
apply(rule rev_mp[OF sym_exec_spec])
apply(rule rev_mp[OF account_def])
apply(rule rev_mp[OF test_purpose])
apply(rule_tac x=x in spec[OF allI])

txt{*  Starting the test generation process.*}
(* Variant I :
using[[no_uniformity]]
apply(gen_test_cases 3 1 "PUT" simp:  HH split: HOL.split_if_asm)
txt{* So lets go for a more non-destructive approach: *}
*)
(* Variant II : *)
apply(gen_test_cases 4 1 "PUT")
(* and symbolic execution separately *)
apply(simp_all add:  HH split: HOL.split_if_asm)

(* test data selection by hand !!! *)
apply(auto intro: P1 P2 P3)
store_test_thm "mykeos_simple"





text{* And now the Fail-Stop scenario --- this corresponds exactly to inclusion test. *}
declare Monads.mbind'_bind [simp del]
test_spec test_status2: 
assumes system_def   : "(c\<^sub>0,no) \<in> dom \<sigma>\<^sub>0" 
and     store_finite  : "\<sigma>\<^sub>0 = map_of T" 
and     test_purpose  : "test_purpose [(c\<^sub>0,no),(c\<^sub>0,no')] S"
and     sym_exec_spec :
       " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return (s = x))"
shows  " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S PUT; return (s = x))"
txt{* Prelude: Massage of the test-theorem --- representing the assumptions of the test explicitely
in HOL and blocking @{term x} from beeing case-splitted (which complicates the process).*}
apply(rule rev_mp[OF sym_exec_spec])
apply(rule rev_mp[OF system_def])
apply(rule rev_mp[OF test_purpose])
apply(rule_tac x=x in spec[OF allI])
txt{*  Starting the test generation process.*}
using[[no_uniformity]] (* bug in uniformity *)
apply(gen_test_cases 3 1 "PUT")
txt{* So lets go for a more non-destructive approach: *}
(* test data selection by hand !!! *)
apply(auto intro:  P1''  P2'')
(* BUG in uniformity-Generation: Separation between Constraints, Programs, 
   and Hyps did not work yet ...
prefer 3
thm spec
apply(tactic "TestGen.uniformityI_tac @{context} [\"PUT\"] 1") *)
store_test_thm "mykeos_simple'"


(* TO DO NEXT: SOLVE Problem with sigma 0. *)
(* TO DO OVERNEXT: Inclusion Test Scenario. *)
thm mykeos_simple.test_thm

subsection{* Rudimentary Test-Data Generation *}

declare [[testgen_iterations=0]]
gen_test_data "mykeos_simple"
thm mykeos_simple.test_data(3)

subsection{*Rudimentary ML-Interface for Test-Sequence*}


ML{*

val test_facts = @{thms mykeos_simple.test_data};
val tfact = nth(test_facts, 2);
 *}

(* Constructors and Destructors for common patterns of Test-Statements *)

ML{*

val (Const(@{const_name "valid_SE"},_) $ _) 
     $  (Const(@{const_name "bind_SE"},_) $ T $ R) = HOLogic.dest_Trueprop(concl_of tfact);
val Const (@{const_name "mbind"}, _) $ LIST $ _= T; 
val S = HOLogic.dest_list LIST;

fun dest_obsTestSeq ((Const(@{const_name "valid_SE"},_) $ state) 
                     $  (Const(@{const_name "bind_SE"},_) $ T 
                          $  Abs (s, ty, Const(@{const_name "unit_SE"}, _) $ P))) =
                     case T of 
                         Const (@{const_name "mbind"}, _) $ L $ SYS => 
                             (state, 0, L, SYS, Abs (s, ty, P))
                       | Const (@{const_name "mbind'"}, _) $ L $ SYS => 
                             (state, 1, L, SYS, Abs (s, ty, P)) 
                       | Const (@{const_name "mbind''"}, _) $ L $ SYS => 
                             (state, 2, L, SYS, Abs (s, ty, P))

 *}

subsection{*Generation of a gdb file*}
 ML{*     
          val file_path_MykeOS = Path.explode "/MyKeOS/impl/c/MykeOS.gdb";
          val file_path_MykeOS = Path.explode "examples/sequence/MykeOS/impl/c/MyKeOS.gdb";

         val file_check_MykeOS = File.exists file_path_MykeOS;

*}

ML {*   
         val term_x = hd S;
       fun actTerm_to_gdbTerm (Const(@{const_name "alloc"}, typ) $ _ $ B $ _)= 
        gdb_header $ gdb_break_point_entry (Const(@{const_name "alloc"}, typ)) B $ 
        gdb_break_point_exist @{term "0"} B $ @{term "start"}
      | actTerm_to_gdbTerm (Const(@{const_name "release"}, typ) $ _ $ B $ _)= 
        gdb_header $ gdb_break_point_entry (Const(@{const_name "release"}, typ)) B $
        gdb_break_point_exist @{term "0"} B $ @{term "start"}
      | actTerm_to_gdbTerm (Const(@{const_name "status"}, typ) $ _ $ B)= 
        gdb_header $ gdb_break_point_entry (Const(@{const_name "alloc"}, typ)) B $
        gdb_break_point_exist @{term "0"} B $ @{term "start"}
     | actTerm_to_gdbTerm (Const(@{const_name "end"}, typ) $ _)= 
        @{term "start"};
   
     val save_it =(render_term @{context} o actTerm_to_gdbTerm) term_x;

     File.write file_path_MykeOS ((String.implode o next_line o remove_char o String.explode) save_it);
      
         *}

ML{* datatype in_event = alloc   of int * int * int
                       | release of int * int * int
                       | status  of int * int ;
val (Const(@{const_name "alloc"}, _) $ A $ B $ C) :: _ = S;

(HOLogic.dest_nat(C));

fun dest_in_event (Const(@{const_name "alloc"}, _) $ A $ B $ C) = 
                  alloc(snd(HOLogic.dest_number(A)),
                        snd(HOLogic.dest_number(B)),
                        HOLogic.dest_nat(C))
   |dest_in_event (Const(@{const_name "release"}, _) $ A $ B $ C) = 
                  release(snd(HOLogic.dest_number(A)),
                        snd(HOLogic.dest_number(B)),
                        HOLogic.dest_nat(C))
   |dest_in_event (Const(@{const_name "status"}, _) $ A $ B) = 
                  status(snd(HOLogic.dest_number(A)),
                        snd(HOLogic.dest_number(B)))
   
val X = map dest_in_event S; 
A
*}

   
declare [[testgen_iterations=0]]
gen_test_data "mykeos_simple'"
thm mykeos_simple'.test_data


subsection {* Generating the Test-Driver for an SML and C implementation *}

text{* The generation of the test-driver is non-trivial in this exercise since it is
essentially two-staged: Firstly, we chose to generate an SML test-driver, which is then
secondly, compiled to a C program that is linked to the actual program under test.
Recall that a test-driver consists of four components:
\begin{itemize}
\item \verb+../../../../../harness/sml/main.sml+ the global controller 
      (a fixed element in the library),
\item \verb+../../../../../harness/sml/main.sml+ a statistic evaluation library
      (a fixed element in the library),   
\item \verb+bank_simple_test_script.sml+ the test-script that corresponds merely
      one-to-one to the generated test-data (generated)
\item \verb+bank_adapter.sml+ a hand-written program; in our scenario, it replaces
      the usual (black-box) program under test by SML code, that calls the 
      external C-functions  via a foreign-language interface.
\end{itemize}
*}

text{* On all three levels, the HOL-level, the SML-level, and the C-level, there are
different representations of basic data-types possible; the translation process of data 
to and from the C-code under test has therefore to be carefully designed 
(and the sheer space of options is sometimes a pain in the neck).
Integers, for example, are represented in two ways inside Isabelle/HOL; there is the mathematical
quotient construction and a "numerals" representation providing 'bit-string-representation-behind-
the-scene" enabling relatively efficient symbolic computation. Both representations can be
compiled "natively" to data types in the SML level. By an appropriate configuration, the 
code-generator can map "int" of HOL to three different implementations:
the SML standard library \verb+Int.int+, the native-C interfaced by \verb+Int32.int+, and the
\verb+IntInf.int+ from the multi-precision library \verb+gmp+ underneath the polyml-compiler.
*}

text{* We do a three-step compilation of data-reresentations model-to-model, model-to-SML, 
SML-to-C. *}

text{* A basic preparatory step for the initializing the test-environment to enable
code-generation is: *}
generate_test_script "mykeos_simple"
thm                   mykeos_simple.test_script

text{* In the following, we describe the interface of the SML-program under test, which is 
in our scenario an \emph{adapter} to the C code under test. This is the heart of the
model-to-SML translation.
The the SML-level stubs for the program under test are declared as follows: *}
consts     status_stub :: "task_id \<Rightarrow> int \<Rightarrow> (int, '\<sigma>)MON\<^sub>S\<^sub>E"
code_const status_stub    (SML "MyKeOSAdapter.status")
consts     alloc_stub  :: "task_id \<Rightarrow> int \<Rightarrow> int \<Rightarrow> (unit, '\<sigma>)MON\<^sub>S\<^sub>E"
code_const alloc_stub     (SML "MyKeOSAdapter.alloc")
consts     release_stub:: "task_id \<Rightarrow> int \<Rightarrow> int \<Rightarrow> (unit, '\<sigma>)MON\<^sub>S\<^sub>E"
code_const release_stub   (SML "MyKeOSAdapter.release")

text{*Note that this translation step prepares already the data-adaption; the type \verb+nat+
is seen as an predicative constraint on integer (which is actually not tested).
On the model-to-model level, we provide a global step function that distributes to
individual interface functions via stubs (mapped via the code generation to SML ...).
This translation also represents uniformly nat by int's.
*}

(* hack : codegenerator does not work for "nat" now ! Therefore an own conversion function:*)
fun    my_nat_conv :: "int \<Rightarrow> nat"
where "my_nat_conv x =(if x <= 0 then 0 else Suc (my_nat_conv(x - 1)))"   

fun   stepAdapter :: "(in_c \<Rightarrow>(out_c, '\<sigma>)MON\<^sub>S\<^sub>E)" 
where
     "stepAdapter(status tid thid) = 
                 (x \<leftarrow> status_stub tid thid; return(status_ok (my_nat_conv x)))"
   | "stepAdapter(alloc tid thid amount) = 
                 (_ \<leftarrow> alloc_stub thid thid (int amount); return(alloc_ok))"
   | "stepAdapter(release tid thid amount)= 
                 (_ \<leftarrow> release_stub tid thid (int amount); return(release_ok))"

text{* The @{term stepAdapter} function links the HOL-world and establishes the logical link
       to HOL stubs which were mapped by the code-generator to adapter functions in SML (which call 
       internally to C-code inside \verb+bank_adapter.sml+ via a foreign language interface) 
       ... We configure the code-generator to identify the \verb+PUT+ with the generated SML code 
       implicitely defined by the above @{term stepAdapter} definition. *}
 
code_const PUT  (SML    "stepAdapter")

text{* And there we go and generate the \verb+mykeos_simple_test_script.sml+: *}
export_code                   stepAdapter mykeos_simple.test_script in SML    
module_name TestScript file  "impl/c/mykeos_simple_test_script.sml"





subsection{* More advanced Test-Case Generation Scenarios *}

text{* Exploring a bit the limits ... *}

text{* Rewriting based approach of symbolic execution ... FailSave Scenario *}
test_spec test_status:
assumes account_def   : "(c\<^sub>0,no) \<in> dom \<sigma>\<^sub>0" 
and     store_finite  : "\<sigma>\<^sub>0 = map_of T" 
and     test_purpose  : "test_purpose [(c\<^sub>0,no),(c\<^sub>0,no')] S"
and     sym_exec_spec :
       " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>a\<^sub>v\<^sub>e S SYS; return (s = x))"
shows  " \<sigma>\<^sub>0  \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>a\<^sub>v\<^sub>e S PUT; return (s = x))"
txt{* Prelude: Massage of the test-theorem --- representing the assumptions of the test explicitely
in HOL and blocking @{term x} from beeing case-splitted (which complicates the process).*}
apply(insert (* store_finite *)  account_def test_purpose sym_exec_spec)
apply(tactic "TestGen.mp_fy 1",rule_tac x=x in spec[OF allI])
txt{*  Starting the test generation process.*}
apply(gen_test_cases 3 1 "PUT")
(* Measurements 20.2. Revision 11395
   5 : 1 =>  31 test cases in 7.609 seconds
   6 : 1 =>  63 test cases in 26.977 seconds 
   7 : 1 => 127 test cases in 125.183 seconds
*)
txt{* Symbolic Execution: *}
apply(simp_all add:  HH split: HOL.split_if_asm)
(* test data selection by hand !!! *)
apply(auto intro: P1' P1 P2 P3)
store_test_thm "mykeos_large"

gen_test_data "mykeos_large"
thm mykeos_large.test_data


text{* Rewriting based approach of symbolic execution ... FailSave Scenario *}
test_spec test_status:
assumes account_def   : "(c\<^sub>0,no) \<in> dom \<sigma>\<^sub>0" 
and     store_finite  : "\<sigma>\<^sub>0 = map_of T" 
and     test_purpose  : "test_purpose [(c\<^sub>0,no),(c\<^sub>0,no')] S"
and     sym_exec_spec : 
       " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return (s = x))"
shows  " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S PUT; return (s = x))"
txt{* Prelude: Massage of the test-theorem --- representing the assumptions of the test explicitely
in HOL and blocking @{term x} from beeing case-splitted (which complicates the process).*}
apply(insert (* store_finite *)  account_def test_purpose sym_exec_spec)
apply(tactic "TestGen.mp_fy 1",rule_tac x=x in spec[OF allI])
txt{*  Starting the test generation process.*}
using [[no_uniformity]] (* To circumvent BUG in uniformityI *)
apply(gen_test_cases 3 1 "PUT")
(* Measurements 20.2. Revision 11395
   5 : 1 => 31 test cases in 3.704 seconds
   6 : 1 => 63 test cases in 5.726 seconds
   7 : 1 => 127 test cases in 25.700 seconds
   8 : 1 => 255 test cases in 119.720 seconds
   9 : 1 => 511 test cases in 755.510 seconds
  10 : 1 => 1023 test cases in 6505.396 seconds
*)
txt{* Symbolic Execution: *}
apply(simp_all add:  HH split: HOL.split_if_asm)
(* test data selection by hand !!! *)
apply(auto intro: P1' P1 P2 P3)
store_test_thm "mykeos_large'"

gen_test_data "mykeos_large'"
thm mykeos_large'.test_data


text{* And now, to compare, elimination based procedures ... *}

declare alloc.exec_mbindFSave_If[simp del]
declare status.exec_mbindFSave_If[simp del]
declare release.exec_mbindFSave_If[simp del]
declare alloc.exec_mbindFStop [simp del]
declare status.exec_mbindFStop[simp del]
declare release.exec_mbindFStop[simp del]

thm alloc.exec_mbindFSave_E release.exec_mbindFSave_E status.exec_mbindFSave_E
(*
declare exec_mbindFSave_allocE[elim!]
declare exec_mbindFSave_releaseE[elim!]
declare exec_mbindFSave_statusE[elim!]
*)


lemma eq2E : "A = B \<Longrightarrow> A \<Longrightarrow> (B \<Longrightarrow> Q) \<Longrightarrow> Q " by auto
lemma valid_mbind_mt : "(\<sigma> \<Turnstile> ( s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p [] f; unit\<^sub>S\<^sub>E (P s))) = P [] " by simp
lemmas H1 = valid_mbind_mt[THEN eq2E]

test_spec test_status:
assumes account_defined: "(c\<^sub>0,no) \<in> dom \<sigma>\<^sub>0" 
and     store_finite   : "\<sigma>\<^sub>0 = map_of T" 
and     test_purpose   : "test_purpose [(c\<^sub>0,no),(c\<^sub>0,no')] S"
and     sym_exec_spec :
       " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return (s = x))"
shows  " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S PUT; return (s = x))"  
apply(insert (* store_finite *)  account_defined test_purpose sym_exec_spec)
apply(tactic "TestGen.mp_fy 1",rule_tac x=x in spec[OF allI])
using [[no_uniformity]]
apply(gen_test_cases 3 1 "PUT" )
(* Measurements 20.2. Revision 11395
   5 : 1 =>   31 test cases in 0.423 seconds
   6 : 1 =>   63 test cases in 1.319 seconds
   7 : 1 =>  127 test cases in 4.626 seconds
   8 : 1 =>  255 test cases in 17.063 seconds
   9 : 1 =>  511 test cases in 86.938 seconds
  10 : 1 =>  1023 test cases in 409.183 seconds
*)
apply(tactic "ALLGOALS(TestGen.REPEAT'(ematch_tac [@{thm status.exec_mbindFStop_E},
                                                   @{thm release.exec_mbindFStop_E},
                                                   @{thm alloc.exec_mbindFStop_E},
                                                   @{thm H1}
                          ]))")
apply(simp_all)
store_test_thm "bank_large_very" 

(* yet another idea: fuse split and symbolic execution: *)
ML{*  Goal.PARALLEL_GOALS *}

theorem hulk : 
assumes redex    : "\<sigma> \<Turnstile> (s \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (a # S) SYS); return (P s))"
and case_alloc : " \<And>c no m. a = alloc c no m \<Longrightarrow> (c, no) \<in> dom \<sigma> \<Longrightarrow> 
                               \<sigma>((c, no) \<mapsto> the (\<sigma> (c, no)) + int m) \<Turnstile>
                                      ( s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return P (alloc_ok # s)) \<Longrightarrow>
                               Q"
and case_release : "\<And>c no m.  a = release c no m \<Longrightarrow> (c, no) \<in> dom \<sigma> \<Longrightarrow>
                               int m \<le> the (\<sigma> (c, no)) \<Longrightarrow>
                               \<sigma>((c,no) \<mapsto> the(\<sigma>(c,no))- int m) \<Turnstile>
                                              (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return P (release_ok#s)) \<Longrightarrow>
                               Q"
and case_status  : "\<And>c no.    (c, no) \<in> dom \<sigma> \<Longrightarrow> 
                                 \<sigma> \<Turnstile>(s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; 
                                     return P (status_ok (nat (the (\<sigma> (c, no)))) # s)) \<Longrightarrow>
                                Q"
shows "Q"
proof(cases "a") print_cases
     case (alloc c no m) assume hyp : "a = alloc c no m" show "Q"
        using hyp redex
        apply(simp only: alloc.exec_mbindFStop) 
        apply(rule case_alloc, auto)
        done
next 
     case (release c no m) assume hyp : "a = release c no m" show "Q"
        using hyp redex
        apply(simp only: release.exec_mbindFStop) 
        apply(rule case_release, auto)
        done
next 
     case (status c no) assume hyp : "a = status c no " show "Q"
        using hyp redex
        apply(simp only: status.exec_mbindFStop) 
        apply(rule case_status, auto)
        done
qed

section{* Partial Order Reduction Approach *}


text{* The continuation monad $M$ is conceptually superfluous; it is used 
       to carry the result type-parameter @{typ "'o'"} to maintain full
       generality *}
definition po_is_exec_consistent :: 
           "('\<iota> \<Rightarrow> ('o,'\<sigma>)MON_SE) \<Rightarrow> ('\<iota> \<Rightarrow> '\<iota> \<Rightarrow>  bool) \<Rightarrow>  ('o \<Rightarrow> 'o \<Rightarrow> ('o','\<sigma>)MON_SE) \<Rightarrow> bool"
where "po_is_exec_consistent iostep ord M=            
          (\<forall> \<iota> \<iota>'. 
            (ord \<iota>' \<iota>) \<longrightarrow>                
             ((o1 \<leftarrow> iostep \<iota>;  (o2 \<leftarrow> iostep \<iota>'; M o1 o2))) = 
             ((o2 \<leftarrow> iostep \<iota>'; (o1 \<leftarrow> iostep \<iota>;  M o1 o2))))"


term "sort x"
fun    ins :: "('a::order) \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "ins x [] = [x]"
     |"ins x (y#ys) = (if (y < x) then y#(ins x ys) else x#y#ys )"
fun    sort:: "('a::order) list \<Rightarrow> 'a list"
where "sort [] = []"
     |"sort (x#xs) = ins x (sort xs)"

lemma ins_nonempty: "ins x S \<noteq> []"
by(induct "S", simp_all)

theorem ordered_exec_state_invariant :
assumes A: "\<forall>M. po_is_exec_consistent iostep (op <) M"
shows      "(\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (sort S) iostep); assert (P))) = 
            (\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S) iostep); assert (P)))" 
proof (induct S)print_cases
   case Nil show ?case by simp
next
   case (Cons a S) from Cons.hyps show ?case 
      proof(cases S) print_binds
         case Nil 
            show "S = [] \<Longrightarrow> 
                    (\<sigma> \<Turnstile> ( _ \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (MyKeOS.sort (a # S)) iostep; assert P)) 
                  = (\<sigma> \<Turnstile> ( _ \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (a # S) iostep; assert P))" 
            by simp
      next
         case (Cons aa list) 
            show "S = aa # list \<Longrightarrow> 
                    (\<sigma> \<Turnstile> ( _ \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (MyKeOS.sort (a # S)) iostep; assert P)) 
                  = (\<sigma> \<Turnstile> ( _ \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (a # S) iostep; assert P))"
            apply(insert ins_nonempty[of aa "MyKeOS.sort list"])
            apply(simp add: List.neq_Nil_conv, elim exE, simp, intro conjI)

      sorry
qed

section{* Commutativities for Re-Orderding Theorems and Partial Order Reduction*}

instantiation in_c::ord
begin

fun  less_in_c where     
    "(alloc c no m)   < (alloc c' no' m')   = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(alloc c no m)   < (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(release c no m) < (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(alloc c no m)   < (status c' no')     = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(release c no m) < (status c' no')     = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(status c no )   < (status c' no' )    = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(X::in_c) < (Y::in_c)  = False"

definition less_eq_in_c_def : "(s::in_c) \<le> t \<longleftrightarrow> s < t \<or> s = t" 

instance ..

end

lemma less_in_c_nonrefl : "(X::in_c) < (Y::in_c) \<Longrightarrow> X \<noteq> Y"
by(cases "X")auto

lemma less_in_c_antisym : "(X::in_c) < (Y::in_c) \<Longrightarrow> \<not>(Y < X)"
apply(cases "X")
apply(cases "Y",auto)+
done

instantiation in_c::order
begin
instance
proof
  fix x::in_c show "x \<le> x"  by(simp add: less_eq_in_c_def) 
next
  fix x::in_c fix y assume "x \<le> y" and "y \<le> x" thus  " x = y"  
      apply(auto simp: less_eq_in_c_def, cases "x")
      apply(cases "y",auto)+
      done
next
  fix x::in_c fix y show  "(x < y) = (x \<le> y \<and> \<not> y \<le> x)"  
      apply(auto simp: less_eq_in_c_def dest:less_in_c_nonrefl)
      apply(drule less_in_c_antisym, simp)
      done
next
  fix x::in_c fix y  fix z show "x \<le> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z" 
      apply(auto simp: less_eq_in_c_def)
      apply(cases "y", simp_all)
      apply(cases "x",cases "z",auto)
      apply(cases "x",cases "z",auto)
      apply(cases "z", auto) apply(cases "z", auto)
      apply(cases "x",cases "z",auto)
      apply(cases "z", auto) +
      done
qed

end


lemma 11:  "((o1 \<leftarrow> SYS (alloc c no m);    (o2 \<leftarrow> SYS (alloc c' no' m');M o1 o2))) = 
            ((o2 \<leftarrow> SYS (alloc c' no' m'); (o1 \<leftarrow> SYS (alloc c no m);   M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done

lemma 12:  "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (alloc c no m);      (o2 \<leftarrow> SYS (release c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (release c' no' m'); (o1 \<leftarrow> SYS (alloc c no m);      M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure release.exec_succes 
                release.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done

lemma 21:  "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (release c no m);  (o2 \<leftarrow> SYS (alloc c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (alloc c' no' m'); (o1 \<leftarrow> SYS (release c no m);  M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure 
                release.exec_succes release.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done


lemma 22: "((o1 \<leftarrow> SYS (release c no m);    (o2 \<leftarrow> SYS (release c' no' m'); M o1 o2))) = 
           ((o2 \<leftarrow> SYS (release c' no' m'); (o1 \<leftarrow> SYS (release c no m);    M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def release.exec_succes release.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_commute)
done


lemma 33: "((o1 \<leftarrow> SYS (status c no);   (o2 \<leftarrow> SYS (status c' no'); M o1 o2))) = 
           ((o2 \<leftarrow> SYS (status c' no'); (o1 \<leftarrow> SYS (status c no);   M o1 o2)))"
by(rule ext,simp add: bind_SE_def status.exec_succes status.exec_failure split:option.split)

lemma 31: "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (status c no);  (o2 \<leftarrow> SYS (alloc c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (alloc c' no' m'); (o1 \<leftarrow> SYS (status c no);  M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure status.exec_succes 
                status.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done

lemma 32:  "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (status c no);  (o2 \<leftarrow> SYS (release c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (release c' no' m'); (o1 \<leftarrow> SYS (status c no);  M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def release.exec_succes release.exec_failure status.exec_succes 
                status.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done


theorem SYS_is_exec_consistent : "po_is_exec_consistent SYS (op <) M"
apply(auto simp: po_is_exec_consistent_def)
apply(rename_tac "\<iota>'" "\<iota>")
apply(case_tac "\<iota>'", auto)
  apply(case_tac "\<iota>", auto, subst 11, simp, subst 11, simp)
 apply(case_tac "\<iota>", auto, subst 21, simp_all,subst 21, simp_all)
 apply(subst 22, simp_all, subst 22, simp_all)
apply(case_tac "\<iota>", auto)
    apply(subst 31, simp_all,subst 31, simp_all)
   apply(subst 32, simp_all,subst 32, simp_all)
  apply(subst 33, simp_all)+
done


section{* Experimental Space *}
declare[[testgen_trace]] 






end